"""
Contains all implementations of history functions.

Section
=======

.. autofunction:: fdSY_mat

Another Section
===============

.. autofunction:: cdSY_mat
"""
from numpy import repeat


# PYTHON PORT ANNOTATION
# Removed parameter K, as it's value can be obtained from
# the shape of x_mat.
def fdSY_mat(mem, x_mat, gradf_mat):
    """Finite difference history ordering.

    Finite difference history ordering for limited memory matrix
    secant methods.

    Parameters
    ----------
    mem : int
        TODO
    x_mat : array_like
        The matrix of iterates.
    gradf_mat : array_like
        The matrix of gradients.

    Returns
    -------
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.

    See Also
    --------
    cdSY_mat : Alternative function.
    history_wrapper : Wrapper to call this function.
    """
    # PYTHON PORT ANNOTATION
    # In the MATLAB code tmp_mat will always contain the mem-th column of
    # x_mat, repeated mem-1 times.
    tmp_mat = repeat(x_mat[:, mem-1], mem-1).reshape((-1, mem-1))
    S_mat = x_mat[:, 0:mem-1] - tmp_mat
    # PYTHON PORT ANNOTATION
    # In the MATLAB code tmp_mat will always contain the mem-th column of
    # gradf_mat, repeated mem-1 times.
    tmp_mat = repeat(gradf_mat[:, mem-1], mem-1).reshape((-1, mem-1))
    Y_mat = gradf_mat[:, 0:mem-1] - tmp_mat

    return S_mat, Y_mat


# PYTHON PORT ANNOTATION
# Removed parameter K, as it's value can be obtained from
# the shape of x_mat.
def cdSY_mat(mem, x_mat, gradf_mat):
    """
    Conjugate direction history ordering.

    Conjugate direction history ordering for limited memory matrix
    secant methods.

    Parameters
    ----------
    mem : int
        TODO
    x_mat : array_like
        The matrix of iterates.
    gradf_mat : array_like
        The matrix of gradients.

    Results
    -------
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.

    See Also
    --------
    fdSY_mat : Alternative function.
    history_wrapper : Wrapper to call this function.
    """
    S_mat = x_mat[:, 1:mem] - x_mat[:, 0:mem-1]
    Y_mat = gradf_mat[:, 1:mem] - gradf_mat[:, 0:mem-1]

    return S_mat, Y_mat


def history_wrapper(History, mem, x_mat, gradf_mat):
    """Wrapper for all history functions.

    Provides a unified function call for all history functions.

    Parameters
    ----------
    mem : int
        TODO
    x_mat : array_like
        The matrix of iterates.
    gradf_mat : array_like
        The matrix of gradients.

    Results
    -------
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.

    See Also
    --------
    fdSY_mat, cdSY_mat
    """
    if History is fdSY_mat:
        return fdSY_mat(mem, x_mat, gradf_mat)

    if History is cdSY_mat:
        return cdSY_mat(mem, x_mat, gradf_mat)



from numpy import arange, c_, dot, load, ndarray, ones, savez, zeros
from numpy.linalg import norm
import matplotlib.pyplot as plt

from .hessian import BFGS1_product, invBFGS1_product,\
    Broyden1_product, invBroyden1_product, Broyden2_product,\
    invBroyden2_product, hessian_wrapper, delta_MSB2,\
    delta_invBFGS, scaler_wrapper
from .history import fdSY_mat, cdSY_mat, history_wrapper
from .memory import Memory_update
from .step_finder import Dogleg_QN, Explicit_TR, Step_length_update,\
    Step_generator


def is_empty(x):
    """Checks if a variable holds any information.

    Determines if `x` is `None` or an empty array.

    Parameters
    ----------
    x
        The variable to be checked.

    Returns
    -------
    bool
        `True` if `x` is either `None` or an empty array. `False` otherwise.
    """
    if x is None:
        return True
    # ndarrays of scalars have a '__len__' attribute, but no len() function.
    if isinstance(x, ndarray):
        if x.size == 0:
            return True
    elif hasattr(x, '__len__') and len(x) == 0:
        return True
    return False


# PYTHON PORT ANNOTATION
# Removed dim and maxmem from parameter list, as they can be obtained by
# reading the dimensions of x_mat.
def Normsqresid(Scaler, History, Hessian_product, HessianT_product,
                mem, spsd, x_mat, gradf_mat,
                gradfnew_vec, alpha_0):
    """Computes the normsquared residual of the gradient or otherwise
    vector-values function.

    Parameters
    ----------
    Scaler : function
        TODO
    History : function
        TODO
    Hessian_product : function
        TODO
    HessianT_product : function
        TODO
    mem : int
        TODO
    spsd : int
        TODO
    x_mat : array_like
        TODO
    gradf_mat : array_like
        TODO
    gradfnew_vec : array_like
        TODO
    alpha_0 : float
        TODO

    Returns
    -------
    f0 : float
        TODO
    gradf0_vec : array_like
        TODO

    See Also
    --------
    delta_invBFGS, delta_MSB2, cdSY_mat, fdSY_mat, BFGS1_product,
    invBFGS1_product, Broyden1_product, invBroyden1_product,
    Broyden2_product, invBroyden2_product
    """
    dim = x_mat.shape[0]
    ######################################################################
    # compute the norm-squared residual
    ######################################################################
    f0 = .5*dot(gradfnew_vec, gradfnew_vec)

    ######################################################################
    # approximate the gradient of the norm-squared residual
    ######################################################################
    # TODO(benedikt) This code is never executed!
    if mem > 0:
        S_mat = zeros((dim, mem))
        Y_mat = zeros((dim, mem))
        S_mat, Y_mat = history_wrapper(mem+1, x_mat[:, -mem-1:-1],
                                       gradf_mat[:, -mem-1:-1])
        delta = scaler_wrapper(Scaler, mem, S_mat, Y_mat, alpha_0)
        alpha = alpha_0
        if spsd >= 0:
            gradf0_vec = gradf_mat[:, -1]
        else:
            gradf0_vec = hessian_wrapper(HessianT_product, mem,
                                         gradf_mat[:, -1], S_mat,
                                         Y_mat, delta, alpha)
            # which means I have to write a butt-load of
            # Hessian_transpose products.
    else:
        # no information
        gradf0_vec = gradfnew_vec

    return f0, gradf0_vec


class Samsara():
    def __init__(self):
        """Sets default values.
        """
        # Declare Parameters
        self.x0_vec = None
        self.xnew_vec = None
        self.f0 = None
        self.fnew = None
        self.gradf0_vec = None
        self.gradfnew_vec = None

        # TODO(benedikt) DELETE ME
        self.debug_file = 'py_vars.npz'
        self.debug = 'step_generator'
        self.TMP = {}
        self.TMP['mem_1'] = -100.
        self.TMP['mem_2'] = -100.
        self.TMP['fnew_1'] = -100.
        self.TMP['gradfnew_1'] = -100.
        self.TMP['tr_slu'] = -100.
        self.TMP['mu_vec_slu'] = -100.
        self.TMP['cDDf_slu'] = -100.
        self.TMP['lower_step_bound_slu'] = -100.
        self.TMP['d_vec_slu'] = -100.
        self.TMP['stepsize_vec_slu'] = -100.
        self.TMP['accept_step_slu'] = -100.
        self.TMP['x_mat_mu'] = -100.
        self.TMP['f_vec_mu'] = -100.
        self.TMP['gradf_mat_mu'] = -100.
        self.TMP['ngradf_vec_mu'] = -100.
        self.TMP['stepsize_vec_mu'] = -100.
        self.TMP['delta_sg'] = -100.
        self.TMP['alpha_sg'] = -100.
        self.TMP['d_vec_sg'] = -100.
        self.TMP['stepsize_vec_sg'] = -100.
        self.TMP['trstep_sg'] = -100.
        self.TMP['DDf_sg'] = -100.
        self.TMP['cDDf_sg'] = -100.
        self.TMP['bDDf_sg'] = -100.
        self.TMP['x_mat_ns'] = -100.
        self.TMP['f_vec_ns'] = -100.
        self.TMP['gradf_mat_ns'] = -100.
        self.TMP['d_vec_ns'] = -100.
        self.TMP['mu_vec_ns'] = -100.
        self.TMP['stepsize_vec_ns'] = -100.
        self.TMP['tr_ns'] = -100.
        self.TMP['delta_ns'] = -100.
        self.TMP['alpha_ns'] = -100.
        self.TMP['eta1_ns'] = -100.
        self.TMP['eta2_ns'] = -100.
        self.TMP['eta3_ns'] = -100.
        self.TMP['cDDf_ns'] = -100.
        self.TMP['DDfnew_ns'] = -100.
        self.TMP['bDDf_ns'] = -100.
        self.TMP['orthog_TOL_ns'] = -100.
        self.TMP['lower_step_bound_ns'] = -100.
        self.TMP['gamma_ns'] = -100.
        self.TMP['d_vec_us'] = -100.
        self.TMP['x_mat_us'] = -100.

        params = {}
        params['verbose'] = True
        params['alpha_0'] = 5e-9
        params['gamma'] = .5
        params['c'] = .01
        params['beta'] = .99
        params['eta1'] = .995
        params['eta2'] = .8
        params['eta3'] = .05
        # params['eta4'] = .025
        params['maxmem'] = 9
        # params['lookback'] = 0
        params['tr'] = 1e+15

        tols = {}
        tols['ngrad_TOL'] = 1e-6
        tols['step_TOL'] = 1e-14
        tols['Maxit'] = 500
        tols['orthog_TOL'] = 1e-6

        opts = {}
        opts['QN_method'] = BFGS1_product
        opts['Step_finder'] = None
        opts['History'] = cdSY_mat
        opts['update_conditions'] = 'Trust Region'
        opts['initial_step'] = 1e+5*tols['step_TOL']

        self.load_options(params, tols, opts)

        self.hist = False

    def __validate_options(self):
        """Validates the set parameters.
        """
        if not (0 < self.c < 1):
            print('The slope modification parameter c in the\
            backtracking subroutine is not in (0,1).')
            print('Setting c to the default, c=0.001.')
            self.c = .001
        if not (0 < self.beta < 1):
            print('The slope modification parameter beta in the\
            backtracking subroutine is not in (0,1).')
            print('Setting beta to the default, beta=0.99.')
            self.beta = .99
        if not (0 < self.gamma < 1):
            print('The backtracking parameter gamma is not in (0,1).')
            print('Setting gamma to the default, gamma=0.5.')
            self.gamma = .5
        if not (self.eta1 > self.eta2 and self.eta1 < 1):
            print('The trust region parameter eta1 is not in (eta2,1).')
            print('Setting to the default, eta1=0.995.')
            self.eta1 = .995
        if not (self.eta2 > self.eta3 and self.eta2 < self.eta1):
            print('The trust region parameter eta2 is not in (eta3,eta1).')
            print('Setting to the default, eta2=0.8.')
            self.eta2 = .8
        if not (self.eta3 > 0 and self.eta3 < self.eta2):
            print('The trust region parameter eta3 is not in (0,eta2).')
            print('Setting to the default, eta3=0.05.')
            self.eta3 = .05
        if self.alpha_0 <= 1e-15:
            print('WARNING: the regularization parameter is probably\
            too small.')
            print('We recommended choosing alpha_0>= 5e-12.')
        if self.step_TOL <= 0:
            raise ValueError('The termination criteria step_TOL sent to the\
            backtracking line search is not positive.',
                             'step_TOL = ' + str(self.step_TOL))

    def __set_methods(self):
        """Sets the methods to use.
        """
        if not (self.History in [cdSY_mat, fdSY_mat]):
            self.History = cdSY_mat
        if self.QN_method is Broyden1_product:
            self.spsd = 0
            self.Hessian_product = Broyden1_product
            self.HessianT_product = Broyden1_product
            self.invHessian_product = invBroyden1_product
            self.Scaler = delta_MSB2
            self.Step_finder = Dogleg_QN
        elif self.QN_method is Broyden2_product:
            self.spsd = 0
            self.Hessian_product = invBroyden2_product
            self.HessianT_product = invBroyden2_product
            self.invHessian_product = Broyden2_product
            self.Scaler = delta_MSB2
            self.Step_finder = Dogleg_QN
        elif self.QN_method is BFGS1_product:
            self.spsd = 1
            self.Hessian_product = BFGS1_product
            self.HessianT_product = BFGS1_product
            self.invHessian_product = invBFGS1_product
            self.Scaler = delta_invBFGS
            if self.Step_finder is None:
                self.Step_finder = Explicit_TR
            if self.History is fdSY_mat:
                print('Finite difference history ordering not consistent\
                with BFGS,')
                print('   changing to conjugate direction history ordering,\
                cdSY_mat.')
                self.History = cdSY_mat
        else:
            # PYTHON PORT ANNOTATION
            # Changed messages to only mention the currently ported methods.
            # Also changed the error message to be more precise.
            print('No such method.  Current options are:')
            print('        Broyden1_product, Broyden2_product, BFGS1_product.')
            raise ValueError('No such QN_method.',
                             'QN_method = ' + str(self.QN_method))

    def __first_step(self):
        """Initialise variables.
        """
        print('First iteration: initialising...')
        ##################################################
        # Initialisation
        ##################################################
        # PYTHON PORT ANNOTATION
        # Removed fcounter, S_mat, Y_mat, as they weren't used.
        self.it = 0   # Python indexing is 0 based
        self.trstep = 0
        self.mem = -1
        # the maxmem column is a temporary storage area
        self.stepsize_vec = zeros(self.Maxit+1)
        ngradf0 = norm(self.gradf0_vec)
        if is_empty(self.gradfnew_vec):
            # automatically generate the first step
            self.d_vec = -self.initial_step*self.gradf0_vec/ngradf0
            self.xnew_vec = self.x0_vec + self.d_vec
            self.gradfnew_vec = 999.*ones(self.dim)
        else:
            self.d_vec = self.xnew_vec - self.x0_vec

        stepsize = norm(self.d_vec)
        self.stepsize_vec[self.it] = stepsize
        self.mu_vec = zeros(self.Maxit+1)
        self.lower_step_bound = 0.
        self.alpha = self.alpha_0
        self.delta = 1.
        self.tr = stepsize
        self.x_mat = zeros((self.dim, self.maxmem+1))
        self.x_mat[:, -2:] = c_[self.x0_vec, self.xnew_vec]
        self.gradf_mat = zeros((self.dim, self.maxmem+1))
        self.gradf_mat[:, -2:] = c_[self.gradf0_vec,
                                    self.gradfnew_vec]
        self.f_vec = zeros(self.Maxit+1)
        if is_empty(self.f0):
            ############################################################
            #   norm squared residual minimization if f0 and fnew empty:
            ############################################################
            self.f0, self.gradf0_vec =\
                Normsqresid(self.Scaler, self.History,
                            self.Hessian_product,
                            self.HessianT_product, self.mem,
                            self.spsd, self.x_mat,
                            self.gradf_mat, self.gradf0_vec,
                            self.alpha_0)
            if is_empty(self.gradfnew_vec):
                # send back to calling driver to get gradfnew
                self.fnew = 999.
            else:
                self.fnew, self.gradfnew =\
                    Normsqresid(self.Scaler, self.History,
                                self.Hessian_product,
                                self.HessianT_product, self.mem,
                                self.spsd, self.x_mat, self.gradf_mat,
                                self.gradfnew_vec, self.alpha_0)
        if is_empty(self.fnew):
            # send back to calling driver to get gradfnew
            self.fnew = 999.

        self.f_vec[self.it] = self.f0
        self.f_vec[self.it+1] = self.fnew
        self.ngradf_vec = zeros(self.Maxit+1)
        self.ngradf_vec[self.it] = ngradf0
        self.ngradf_vec[self.it+1] = norm(self.gradfnew_vec)
        self.DDf = dot(self.gradf_mat[:, -2], self.d_vec)
        self.cDDf = self.c*self.DDf
        self.bDDf = self.beta*self.DDf

    def __prepare_step(self):
        """Get missing function values.
        """
        if is_empty(self.fnew):
            ############################################################
            #   norm squared residual minimization if f0 and fnew empty:
            ############################################################
            self.fnew, self.gradfnew_vec =\
                Normsqresid(self.Scaler, self.History,
                            self.Hessian_product,
                            self.HessianT_product, self.mem,
                            self.spsd, self.x_mat, self.gradf_mat,
                            self.gradfnew_vec, self.alpha_0)
        self.f_vec[self.it+1] = self.fnew
        self.gradf_mat[:, -1] = self.gradfnew_vec
        self.ngradf_vec[self.it+1] = norm(self.gradfnew_vec)
        self.TMP['fnew_1'] = self.fnew
        self.TMP['gradfnew_1'] = self.gradfnew_vec

    def __next_step(self):
        """Compute next point.
        """
        DDfnew = dot(self.gradf_mat[:, -1], self.d_vec)

        self.TMP['x_mat_ns'] = self.x_mat.copy()
        self.TMP['f_vec_ns'] = self.f_vec.copy()
        self.TMP['gradf_mat_ns'] = self.gradf_mat.copy()
        self.TMP['d_vec_ns'] = self.d_vec.copy()
        self.TMP['mu_vec_ns'] = self.mu_vec.copy()
        self.TMP['stepsize_vec_ns'] = self.stepsize_vec.copy()
        self.TMP['tr_ns'] = self.tr
        self.TMP['delta_ns'] = self.delta
        self.TMP['alpha_ns'] = self.alpha
        self.TMP['eta1_ns'] = self.eta1
        self.TMP['eta2_ns'] = self.eta2
        self.TMP['eta3_ns'] = self.eta3
        self.TMP['cDDf_ns'] = self.cDDf
        self.TMP['DDfnew_ns'] = DDfnew
        self.TMP['bDDf_ns'] = self.bDDf
        self.TMP['orthog_TOL_ns'] = self.orthog_TOL
        self.TMP['lower_step_bound_ns'] = self.lower_step_bound
        self.TMP['gamma_ns'] = self.gamma
        #################################################################
        #   ADJUST STEP LENGTH/TRUST REGION and determine Update criteria
        ##################################################################
        self.tr, self.mu_vec, self.cDDf, self.lower_step_bound, self.d_vec,\
            self.stepsize_vec, accept_step =\
            Step_length_update(self.History, self.Hessian_product,
                               self.update_conditions, self.mem,
                               self.it, self.x_mat, self.f_vec,
                               self.gradf_mat, self.d_vec,
                               self.mu_vec, self.stepsize_vec,
                               self.tr, self.delta, self.alpha,
                               self.eta1, self.eta2, self.eta3,
                               self.cDDf, DDfnew, self.bDDf,
                               self.orthog_TOL, self.lower_step_bound,
                               self.gamma)

        self.TMP['tr_slu'] = self.tr
        self.TMP['mu_vec_slu'] = self.mu_vec.copy()
        self.TMP['cDDf_slu'] = self.cDDf
        self.TMP['lower_step_bound_slu'] = self.lower_step_bound
        self.TMP['d_vec_slu'] = self.d_vec.copy()
        self.TMP['stepsize_vec_slu'] = self.stepsize_vec.copy()
        self.TMP['accept_step_slu'] = accept_step

        #####################################################
        #  Update the memory
        #####################################################
        self.x_mat, self.f_vec, self.gradf_mat, self.ngradf_vec,\
            self.stepsize_vec, self.mem =\
            Memory_update(self.x_mat, self.f_vec, self.gradf_mat,
                          self.ngradf_vec, self.stepsize_vec,
                          self.mem, self.spsd, self.it,
                          accept_step)

        self.TMP['mem_1'] = self.mem
        self.TMP['x_mat_mu'] = self.x_mat.copy()
        self.TMP['f_vec_mu'] = self.f_vec.copy()
        self.TMP['gradf_mat_mu'] = self.gradf_mat.copy()
        self.TMP['ngradf_vec_mu'] = self.ngradf_vec.copy()
        self.TMP['stepsize_vec_mu'] = self.stepsize_vec.copy()

        if accept_step > -2:
            # If criteria satisfied, then update and increment iterates
            if accept_step >= 0:
                #####################################################
                # reset lower_step_bound increment it
                #####################################################
                self.lower_step_bound = 0.
                self.it += 1

            #####################################################
            # choose new search direction (and length):
            #####################################################
            self.delta, self.alpha, self.d_vec, self.stepsize_vec,\
                self.trstep, self.DDf, self.cDDf, self.bDDf, self.mem =\
                Step_generator(self.Step_finder, self.Scaler,
                               self.History, self.Hessian_product,
                               self.invHessian_product, self.it,
                               self.mem, self.c, self.beta,
                               self.gamma, self.x_mat, self.gradf_mat,
                               self.ngradf_vec, self.stepsize_vec,
                               self.tr, self.trstep, self.alpha_0,
                               self.verbose)

            self.TMP['mem_2'] = self.mem
            self.TMP['delta_sg'] = self.delta
            self.TMP['alpha_sg'] = self.alpha
            self.TMP['d_vec_sg'] = self.d_vec.copy()
            self.TMP['stepsize_vec_sg'] = self.stepsize_vec.copy()
            self.TMP['trstep_sg'] = self.trstep
            self.TMP['DDf_sg'] = self.DDf
            self.TMP['cDDf_sg'] = self.cDDf
            self.TMP['bDDf_sg'] = self.bDDf

    # TODO(benedikt) Integrate with run()?
    def __update_step(self):
        """Updates the the matrix of steps.
        """
        # update the proposed step
        self.x_mat[:, self.maxmem] = self.x_mat[:, self.maxmem-1] + self.d_vec
        self.TMP['d_vec_us'] = self.d_vec.copy()
        self.TMP['x_mat_us'] = self.x_mat.copy()
        if self.verbose:
            print('iteration:', self.it, '; trust region:',
                  self.tr, '; memory:', self.mem)

        # PYTHON PORT ANNOTATION
        # Inverted condition for improved readabilty.
        if not (self.it < self.Maxit):
            print('Warning:   Maxit specified in OPTIONS_tolerances.m \
            is smaller')
            print('  than the maximum iterations specified in the driver \
            program.')
            print('  Increase the Maxit in OPTIONS_tolerances to at least \
            that of')
            print('  the driver program.')
            # PYTHON PORT ANNOTATION
            # Added an error message.
            raise ValueError('Iterations reached maximum defined in Maxit',
                             'i = ' + str(self.it),
                             'Maxit = ' + str(self.Maxit))

    def load_options(self, params, tols, opts):
        """Set options.
        """
        for k, v in params.items():
            if k == 'verbose':
                self.verbose = v
            if k == 'alpha_0':
                self.alpha_0 = v
            if k == 'gamma':
                self.gamma = v
            if k == 'c':
                self.c = v
            if k == 'beta':
                self.beta = v
            if k == 'eta1':
                self.eta1 = v
            if k == 'eta2':
                self.eta2 = v
            if k == 'eta3':
                self.eta3 = v
            # if k == 'eta4':
            #     self.eta4 = v
            if k == 'maxmem':
                self.maxmem = v
            if k == 'tr':
                self.tr = v

        for k, v in tols.items():
            if k == 'ngrad_TOL':
                self.ngrad_TOL = v
            if k == 'step_TOL':
                self.step_TOL = v
            if k == 'Maxit':
                self.Maxit = v
            if k == 'orthog_TOL':
                self.orthog_TOL = v

        for k, v in opts.items():
            if k == 'QN_method':
                self.QN_method = v
            if k == 'Step_finder':
                self.Step_finder = v
            if k == 'History':
                self.History = v
            if k == 'update_conditions':
                self.update_conditions = v
            if k == 'initial_step':
                self.initial_step = v

        self.__validate_options()
        self.__set_methods()

    def run(self, x0_vec, xnew_vec, f0, fnew, gradf0_vec, gradfnew_vec):
        """Computes the next point. Main routine of the SAMSARA toolbox.

        Parameters
        ----------
        x0_vec : array_like
            TODO
        xnew_vec : array_like
            TODO
        f0 : float
            TODO
        fnew : float
            TODO
        gradf0_vec : array_like
            TODO
        gradfnew_vec : array_like
            TODO

        Returns
        -------
        xnew_vec : array_like
            The new point.
        x0_vec : array_like
            The old point.
        f0 : float
            The old function value.
        gradf0_vec : array_like
            The old gradient vector.
        stepsize : float
            The length of the last step.
        """
        self.dim = x0_vec.shape[0]

        self.x0_vec = x0_vec
        self.xnew_vec = xnew_vec
        self.f0 = f0
        self.fnew = fnew
        self.gradf0_vec = gradf0_vec
        self.gradfnew_vec = gradfnew_vec

        # first iteration
        if not self.hist:
            self.__first_step()
            self.hist = True
        else:
            self.__prepare_step()
            self.__next_step()

        self.__update_step()

        self.xnew_vec = self.x_mat[:, self.maxmem]
        self.x0_vec = self.x_mat[:, self.maxmem-1]
        self.f0 = self.f_vec[self.it]
        self.gradf0_vec = self.gradf_mat[:, self.maxmem-1]
        stepsize = self.stepsize_vec[self.it]

        # TODO(benedikt) DELETE ME
        if self.debug == 'final':
            # final values
            savez(self.debug_file, xnew_vec=self.xnew_vec,
                  x0_vec=self.x0_vec, f0=self.f0,
                  gradf0_vec=self.gradf0_vec, stepsize=stepsize,
                  tr=self.tr, mem=self.mem)
        elif self.debug == 'step_length_update':
            # Step_length_update
            savez(self.debug_file, mu_vec_slu=self.TMP['mu_vec_slu'],
                  cDDf_slu=self.TMP['cDDf_slu'],
                  lower_step_bound_slu=self.TMP['lower_step_bound_slu'],
                  d_vec_slu=self.TMP['d_vec_slu'],
                  stepsize_vec_slu=self.TMP['stepsize_vec_slu'],
                  accept_step_slu=self.TMP['accept_step_slu'],
                  tr_slu=self.TMP['tr_slu'])
        elif self.debug == 'next_step':
            # __next_step beginning
            savez(self.debug_file, x_mat_ns=self.TMP['x_mat_ns'],
                  f_vec_ns=self.TMP['f_vec_ns'],
                  gradf_mat_ns=self.TMP['gradf_mat_ns'],
                  d_vec_ns=self.TMP['d_vec_ns'],
                  mu_vec_ns=self.TMP['mu_vec_ns'],
                  stepsize_vec_ns=self.TMP['stepsize_vec_ns'],
                  tr_ns=self.TMP['tr_ns'],
                  delta_ns=self.TMP['delta_ns'],
                  alpha_ns=self.TMP['alpha_ns'],
                  eta1_ns=self.TMP['eta1_ns'],
                  eta2_ns=self.TMP['eta2_ns'],
                  eta3_ns=self.TMP['eta3_ns'],
                  cDDf_ns=self.TMP['cDDf_ns'],
                  DDfnew_ns=self.TMP['DDfnew_ns'],
                  bDDf_ns=self.TMP['bDDf_ns'],
                  orthog_TOL_ns=self.TMP['orthog_TOL_ns'],
                  lower_step_bound_ns=self.TMP['lower_step_bound_ns'],
                  gamma_ns=self.TMP['gamma_ns'])
        elif self.debug == 'memory_update':
            # Memory_update
            savez(self.debug_file, x_mat_mu=self.TMP['x_mat_mu'],
                  f_vec_mu=self.TMP['f_vec_mu'],
                  gradf_mat_mu=self.TMP['gradf_mat_mu'],
                  ngradf_vec_mu=self.TMP['ngradf_vec_mu'],
                  stepsize_vec_mu=self.TMP['stepsize_vec_mu'],
                  accept_step_slu=self.TMP['accept_step_slu'])
        elif self.debug == 'step_generator':
            # Step_generator
            savez(self.debug_file, delta_sg=self.TMP['delta_sg'],
                  alpha_sg=self.TMP['alpha_sg'],
                  d_vec_sg=self.TMP['d_vec_sg'],
                  stepsize_vec_sg=self.TMP['stepsize_vec_sg'],
                  trstep_sg=self.TMP['trstep_sg'],
                  DDf_sg=self.TMP['DDf_sg'],
                  cDDf_sg=self.TMP['cDDf_sg'],
                  bDDf_sg=self.TMP['bDDf_sg'])
        elif self.debug == 'update_step':
            savez(self.debug_file, d_vec_us=self.TMP['d_vec_us'],
                  x_mat_us=self.TMP['x_mat_us'])
        else:
            # we always need a file for debug_driver
            # savez(self.debug_file, x_mat=self.x_mat)
            pass

        return self.xnew_vec, self.x0_vec, self.f0, self.gradf0_vec, stepsize

    def save(self, filename='state.npz'):
        """Saves some values of samsara.

        Saves `dim`, `f_vec`, `ngradf_vec`, `it`, `stepsize_vec`,
        `mu_vec` to the file indicated by `filename`.

        Parameters
        ----------
        filename : Optional[str]
            The file to write to. Default: `state.npz`
        """
        savez(filename, dim=self.dim, f_vec=self.f_vec,
              ngradf_vec=self.ngradf_vec, it=self.it,
              stepsize_vec=self.stepsize_vec, mu_vec=self.mu_vec)

    def plot(self, objective, state_file=None):
        """Displays a standard plot for `Samsara`.

        If `state_file` is given, then it's values we be used for the
        plot. Otherwise the valuues of the current instance of
        `Samsara` will be used.

        Parameters
        ----------
        objective : function
            The function samsara was minimising.
        state_file : Optional[str]
            A save file created by ``save``.

        See Also
        --------
        save : Used to create the `state_file`.
        """
        if state_file is not None:
            try:
                state = load(state_file)
            except:
                print('No such file:', state_file)
                return
            dim = state['dim']
            f_vec = state['f_vec']
            ngradf_vec = state['ngradf_vec']
            it = state['it']
        elif self.hist:
            dim = self.dim
            f_vec = self.f_vec
            ngradf_vec = self.ngradf_vec
            stepsize_vec = self.stepsize_vec
            it = self.it
        else:
            print('Nothing to plot yet!')
            print('Try calling samsara.run() first.')
            return

        t = arange(it+1)
        f_opt, _ = objective(ones(dim))

        plt.subplot(221)
        f_adj = abs(f_vec[:it+1] - f_opt)
        plt.semilogy(t, f_adj[t])
        plt.title('Function value')
        plt.xlabel('iterations')
        plt.ylabel('f(x)')

        plt.subplot(222)
        plt.semilogy(t, ngradf_vec[t])
        plt.title('gradient norm')
        plt.xlabel('iteration')
        plt.ylabel('log(gradient norm)')

        plt.subplot(223)
        plt.semilogy(t, stepsize_vec[t])
        plt.title('stepsize')
        plt.xlabel('iterations')
        plt.ylabel('||x_{k+1}-x_k||')

        plt.show()



"""

Extending samsara
-----------------
Any functions from additional modules have to be imported in this file.

"""

### MODULE FUNCTION IMPORTS ###

# Samsara
from .samsara import Samsara

# Hessian products and Scaler
from .hessian import BFGS1_product, invBFGS1_product, Broyden1_product,\
    invBroyden1_product, Broyden2_product, invBroyden2_product,\
    delta_MSB2, delta_invBFGS

# History
from .history import fdSY_mat, cdSY_mat

# Step finder
from .step_finder import Dogleg_QN, Explicit_TR



"""Example of a driver for Samsara.

To create a samsara driver for your own project, copy this file and
add in your function and options.
"""
import matplotlib.pyplot as plt
from numpy import abs, amax, array, load, zeros
from numpy.linalg import norm

from samsara import Samsara, BFGS1_product, Broyden1_product,\
    Broyden2_product, Dogleg_QN, Explicit_TR, cdSY_mat, fdSY_mat


def objective(x):
    """Example objective (Engvall function).

    Replace this code with your actual function.

    Parameters
    ----------
    x : array_like
        The point to evaluate the function at (2 dimensional).

    Returns
    -------
    f : float
        The function value at `x`.
    Df : array_like
        The gradient vector of ``f`` at ``x``.

    Notes
    -----
    In your implementation you may use any number of parameters and
    return values. But samsara will always need the function value
    and the gradient vector.
    """
    f = x[0]**4 + x[1]**4 + 2*x[0]**2*x[1]**2 - 4*x[0] + 3
    Df = zeros(2)
    Df[0] = 4*x[0]**3 + 4*x[0]*x[1]**2 - 4
    Df[1] = 4*x[1]**3 + 4*x[0]**2*x[1]

    return f, Df


def plot_result():
    """Example of a custom plot function for the result.

    Replace this code with your actual plotting routine.
    """
    pass


def driver():
    # You can either implement your function in the objective function
    # provided above, or import your function from a different module
    # and set it here.
    Objective = objective

    # ========== SAMSARA ==========
    samsara = Samsara()

    # --- OPTIONS ---
    params = {}
    # Extra output from samsara.
    params['verbose'] = True
    # A regularisation parameter.
    params['alpha_0'] = 5e-12
    # ...
    params['gamma'] = .5
    # ...
    params['c'] = .01
    # ...
    params['beta'] = .9999999
    # ...
    params['eta1'] = .995
    # ...
    params['eta2'] = .8
    # ...
    params['eta3'] = .25
    # The maximum steps to look back. Must be an integer.
    params['maxmem'] = 8
    # The initial radius of the trust region.
    params['tr'] = 1e+15

    tols = {}
    # Gradient norm tolerance
    # This is used as a stopping criteria for SAMSARA. It must be a
    # real number greater than zero.
    tols['ngrad_TOL'] = 1e-6
    # Step size tolerance
    # This is used as a stopping criteria for SAMSARA. It must be a
    # real number greater than zero.
    tols['step_TOL'] = 1e-12
    # Maximum number of iterations
    # This is used as a stopping criteria for SAMSARA. It must be
    # an integer greater than zero.
    tols['Maxit'] = 1000

    opts = {}
    # Hessian method
    # Quasi-Newton methods for estimation of the function's Hessian.
    # Broyden implementations use Dogleg_QN lin search methods
    # while BFGS implementations use Explicit_TR.
    # Methods available:
    #     - BFGS1_product
    #     - Broyden1_product
    #     - Broyden2_product
    opts['QN_method'] = BFGS1_product
    # Step finder
    # Optional: Default is Dogleg_QN
    # Methods available:
    #     - Dogleg_QN
    #     - Explicit_TR
    opts['Step_finder'] = Explicit_TR
    # History
    # Ordering methods for S, Y matrices in limited memory application.
    # Finite difference ordering, fdSY_mat, is recommended for Broyden
    # implementations. For BFGS implementations, use conjugate
    # direction ordering, cdSY_mat.
    opts['History'] = cdSY_mat
    # Trust Region Adjust
    # Select the method for adjustment of the trust region in optimazation.
    opts['update_conditions'] = 'Trust Region'
    # Initial Step Scaler
    # The norm of the initial step taken in the Cauchy direction.
    # This multiplied against the normalized gradient to yield the
    # initial direction vector in order to generate the first step
    # taken by SAMSARA.
    # Assign a value of 0.0D0 to use the default value which is the
    # minimum of 1D-1 or the norm(1d-1*gradient).
    opts['initial_step'] = 1e+5*tols['step_TOL']

    samsara.load_options(params, tols, opts)

    # ========== STARTING VALUES ==========
    # The starting vector. Set it to an appropriate value.
    xold_vec = array([.5, 2.])
    # The starting function value and gradient vector.
    fold, gradfold_vec = Objective(xold_vec)

    # The next point.
    xnew_vec = None
    # The next function value.
    fnew = None
    # The next gradient vector.
    gradfnew_vec = None
    # Keep at this value, so that the main loop executes at least once.
    stepsize = 999.
    # Keep at this value, so that the main loop executes at least once.
    ngradfnew = 999.

    # ========== MAIN LOOP ==========
    # --- Tolerances ---
    # The lower bound for the gradient vector norm.
    ngrad_TOL = 2e-14
    # The lower bound for the stepsize.
    step_TOL = 2e-17
    # The maximum number of iterations.
    Maxit = 500
    # Number of current iterations
    it = 0

    while it < Maxit and ngradfnew > ngrad_TOL and stepsize > step_TOL:
        xnew_vec, xold_vec, fold, gradfold_vec, stepsize =\
            samsara.run(xold_vec, xnew_vec, fold, fnew, gradfold_vec,
                        gradfnew_vec)
        it += 1
        fnew, gradfnew_vec = Objective(xnew_vec)
        ngradfnew = norm(gradfnew_vec)

    if stepsize <= step_TOL:
        print('Algorithm stagnated:  stepsize tolerance violated.')
    if ngradfnew <= ngrad_TOL:
        print('Algorithm stagnated:  gradient norm tolerance violated.')
    if it >= Maxit:
        print('Algorithm exceeded:  maximum step count violated.')

    # TODO(benedikt) Rework from here
    # You can save some of the values of samsara.
    # samsara.save()

    # You can now plot your results.
    # plot_result()

    # Return the optimal values.
    return xnew_vec, fnew


# Entry point into the driver.
if __name__ == '__main__':
    driver()



"""Contains all implementations of hessian products.

Hessian functions
=================

Some introductory text.

.. autofunction:: samsara.hessian.BFGS1_product
.. autofunction:: samsara.hessian.Broyden1_product
.. autofunction:: samsara.hessian.Broyden2_product
.. autofunction:: samsara.hessian.invBFGS1_product
.. autofunction:: samsara.hessian.invBroyden1_product
.. autofunction:: samsara.hessian.invBroyden2_product
.. autofunction:: samsara.hessian_wrapper


Scaler functions
================

These functions are used to determine the `scale` variable for the
hessian functions.

.. autofunction:: samsara.hessian.delta_invBFGS
.. autofunction:: samsara.hessian.delta_MSB2
.. autofunction:: samsara.hessian.scaler_wrapper

"""
from numpy import abs, c_, dot, diag, eye, r_, sqrt, tril, triu, zeros
from numpy.linalg import cholesky, norm, solve
from numpy.random import random, seed


# PYTHON PORT ANNOTATION
# Removed parameters dim and mem, as the value can be obtained from
# the shape of S_mat/Y_mat.
def delta_MSB2(S_mat, Y_mat, alpha):
    """Scaling for the MSB2 matrix.

    Parameters
    ----------
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.
    alpha : float
        A regularisation parameter.

    Returns
    -------
    float
        TODO

    See Also
    --------
    delta_invBFGS : Alternative function.
    samsara.samsara.Normsqresid, samsara.step_finder.Step_generator
    """
    dim = Y_mat.shape[1]
    # normalize the columns of Y_mat for stability
    ny = zeros(dim)
    for j in range(dim):
        ny[j] = 1./norm(Y_mat[:, j])

    # PYTHON PORT ANNOTATION
    # In Python we can't modify the input parameters as they are
    # just references to the originals in the calling function.
    Yc_mat = dot(Y_mat, diag(ny))
    Sc_mat = dot(S_mat, diag(ny))

    # Look into doing a Cholesky factorization of Y'Y for stability
    M_mat = solve(dot(Yc_mat.T, Yc_mat) + alpha*eye(dim, dim), Yc_mat.T)
    seed(5489)   # sets the same seed for the same 'random' vector
    v0_vec = random(S_mat.shape[0])   # used to compute Rayleigh quotient

    delta0 = 0
    delta = 10

    RQ_iter = 0
    v_vec = dot(Sc_mat, dot(M_mat, v0_vec))
    while abs((delta - delta0)/delta) > 1e-5 and RQ_iter < 10:
        RQ_iter += 1
        v0_vec = v_vec/sqrt(dot(v_vec, v_vec))
        delta0 = delta
        v_vec = dot(Sc_mat, dot(M_mat, v0_vec))
        delta = dot(v0_vec, v_vec)

    return abs(delta)


# PYTHON PORT ANNOTATION
# Removed parameter dim, as the value can be obtained from
# the shape of S_mat/Y_mat.
def delta_invBFGS(mem, S_mat, Y_mat):
    """Scaling for the inverse BFGS matrix

    Parameters
    ----------
    mem : int
        The step to use.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.

    Returns
    -------
    float
        TODO

    See Also
    --------
    BFGS1_product : TODO
    delta_MSB2 : Alternative function.
    """
    # print('--- in delta_invBFGS:58 ---')
    # print('S_mat[:dim, mem-1] =', S_mat[:dim, mem-1])
    # print('Y_mat[:dim, mem-1] =', Y_mat[:dim, mem-1])
    # print('--- out delta_invBFGS:61 ---')
    return abs(dot(S_mat[:, mem-1], Y_mat[:, mem-1]))\
        / dot(Y_mat[:, mem-1], Y_mat[:, mem-1])


def scaler_wrapper(Scaler, mem, S_mat, Y_mat, alpha):
    """A wrapper for all scaler functions.

    Used to have unified function calls.

    Parameters
    ----------
    Scaler : function
        The scaler function to use.
    mem : int
        The step to use.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.

    Returns
    -------
    float
        TODO

    See Also
    --------
    delta_MSB2, delta_invBFGS
    """
    if Scaler is delta_invBFGS:
        return delta_invBFGS(mem, S_mat, Y_mat)

    if Scaler is delta_MSB2:
        return delta_MSB2(S_mat, Y_mat, alpha)


def BFGS1_product(u_vec, S_mat, Y_mat, scale):
    """An implementation of the BFGS algorithm.

    Parameters
    ----------
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradient differences.
    scale : float
        The initial scaling.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    delta_invBFGS : TODO
    Broyden1_product, Broyden2_product
    """
    STY_mat = dot(S_mat.T, Y_mat)
    D_mat = diag(diag(STY_mat))
    L_mat = tril(STY_mat, -1)

    Di_mat = diag(diag(STY_mat)**(-1))
    R_mat = 1./scale*dot(S_mat.T, S_mat) + dot(L_mat, dot(Di_mat, L_mat.T))
    R_mat = cholesky(R_mat).T

    Dsqrt_mat = diag(diag(STY_mat)**.5)
    Disqrt_mat = diag(diag(STY_mat)**(-.5))
    Phi_mat = c_[Y_mat, (1./scale)*S_mat]
    Gam1_mat = r_[c_[-Dsqrt_mat, dot(Disqrt_mat, L_mat.T)],
                  c_[zeros(D_mat.shape), R_mat]]
    Gam2_mat = r_[c_[Dsqrt_mat, zeros(D_mat.shape)],
                  c_[-dot(L_mat, Disqrt_mat), R_mat.T]]

    v_vec = dot(Phi_mat.T, u_vec)
    v_vec = solve(Gam2_mat, v_vec)
    v_vec = solve(Gam1_mat, v_vec)
    v_vec = (1./scale)*u_vec - dot(Phi_mat, v_vec)

    return v_vec


def invBFGS1_product(u_vec, S_mat, Y_mat, scale):
    """Stable inverse-BFGS matrix product.

    Parameters
    ----------
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradient differences.
    scale : float
        The initial scaling.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    invBroyden1_product, invBroyden2_product
    """
    # Prevent problems with oriented vectors
    U_vec = u_vec.flatten()
    STY_mat = dot(S_mat.T, Y_mat)
    D_mat = diag(diag(STY_mat))
    R_mat = triu(STY_mat + 1e-11*eye(STY_mat.shape[0], STY_mat.shape[1]), 0)
    Rt_mat = R_mat.T

    v1_vec = dot(S_mat.T, U_vec)
    v2_vec = scale * dot(Y_mat.T, U_vec)
    v3_vec = solve(R_mat, v1_vec)
    v3_vec = dot(D_mat + scale*dot(Y_mat.T, Y_mat), v3_vec) - v2_vec
    v3_vec = solve(Rt_mat, v3_vec)
    v2_vec = -solve(R_mat, v1_vec)
    v_vec = scale*U_vec + dot(S_mat, v3_vec) + scale*dot(Y_mat, v2_vec)

    return v_vec


def Broyden1_product(mem, u_vec, S_mat, Y_mat, scale):
    """An implementation of Broyden's method.

    Parameters
    ----------
    mem : int
        TODO
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradient differences.
    scale : float
        The initial scaling.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    Broyden2_product, BFGS1_product
    """
    tmpzw = dot(S_mat[:, mem-1], u_vec) / dot(S_mat[:, mem-1], S_mat[:, mem-1])
    # Ensure, that Wtmp_vec.shape = (x,) and not (x,1). In the latter case
    # we would have to transpose Wtmp_vec in every addition later on to not
    # suddenly get a matrix as a result of the operation.
    Wtmp_vec = u_vec.flatten()
    Ztmp_vec = tmpzw*Y_mat[:, mem-1]

    # PYTHON PORT ANNOTATION
    # Changed the order of the loop for improved readability.
    for j in range(mem-1, 0, -1):
        Wtmp_vec = Wtmp_vec - tmpzw*S_mat[:, j]
        tmpzw = dot(S_mat[:, j-1], Wtmp_vec) / dot(S_mat[:, j-1],
                                                   S_mat[:, j-1])
        Ztmp_vec = Ztmp_vec + tmpzw*Y_mat[:, j-1]

    Wtmp_vec = Wtmp_vec - tmpzw*S_mat[:, 0]
    v_vec = scale*Wtmp_vec + Ztmp_vec

    return v_vec


def invBroyden1_product(u_vec, S_mat, Y_mat, scale, alpha):
    """An implementation of the inverse Broyden's method.

    Parameters
    ----------
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradient differences.
    scale : float
        The initial scaling.
    alpha : float
        A regularization parameter.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    invBFGS1_product, invBroyden2_product
    """
    dim = Y_mat.shape[1]
    # Prevent problems with oriented vectors
    U_vec = u_vec.flatten()
    M_mat = -tril(dot(S_mat.T, S_mat), -1)
    v_vec = solve(M_mat + scale*dot(S_mat.T, Y_mat) + alpha*eye(dim, dim),
                  scale*dot(S_mat.T, U_vec))
    v_vec = scale*U_vec - dot(scale*Y_mat - S_mat, v_vec)

    return v_vec


def Broyden2_product(mem, u_vec, S_mat, Y_mat, scale):
    """An implementation of Broyden's method.

    Parameters
    ----------
    mem : int
        How many old steps to consider.
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradient differences.
    scale : float
        The initial scaling.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    Broyden1_product, BFGS1_product
    """
    tmpzw = dot(Y_mat[:, mem-1], u_vec) / dot(Y_mat[:, mem-1], Y_mat[:, mem-1])
    # Ensure, that Wtmp_vec.shape = (x,) and not (x,1). In the latter case
    # we would have to transpose Wtmp_vec in every addition later on to not
    # suddenly get a matrix as a result of the operation.
    Wtmp_vec = u_vec.flatten()
    Ztmp_vec = tmpzw*S_mat[:, mem-1]

    # PYTHON PORT ANNOTATION
    # Changed the order of the loop for improved readability.
    for j in range(mem-1, 0, -1):
        Wtmp_vec = Wtmp_vec - tmpzw*Y_mat[:, j]
        tmpzw = dot(Y_mat[:, j-1], Wtmp_vec) / dot(Y_mat[:, j-1],
                                                   Y_mat[:, j-1])
        Ztmp_vec = Ztmp_vec + tmpzw*S_mat[:, j-1]

    Wtmp_vec = Wtmp_vec - tmpzw*Y_mat[:, 0]
    v_vec = scale*Wtmp_vec + Ztmp_vec

    return v_vec


def invBroyden2_product(u_vec, S_mat, Y_mat, scale, alpha):
    """An implementation of the inverse Broyden's method.

    Parameters
    ----------
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradient differences.
    scale : float
        The initial scaling.
    alpha : float
        A regularization parameter.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    invBFGS1_product, invBroyden1_product
    """
    dim = Y_mat.shape[1]
    # Prevent problems with oriented vectors
    U_vec = u_vec.flatten()
    M_mat = -tril(dot(Y_mat.T, Y_mat), -1)
    v_vec = solve(M_mat + 1./scale*dot(Y_mat.T, S_mat) + alpha*eye(dim, dim),
                  1./scale*dot(Y_mat.T, U_vec))
    v_vec = 1./scale*U_vec - dot(1./scale*S_mat - Y_mat, v_vec)

    return v_vec


def hessian_wrapper(Hessian_product, mem, u_vec, S_mat, Y_mat,
                    scale, alpha):
    """Wrapper for all hessian functions in this module.

    Used to have unified parameters across all function calls.

    Parameters
    ----------
    Hessian_product : function
        The hessian function to use.
    mem : int
        The step to use.
    u_vec : array_like
        The vector to be multiplied.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.
    scale : float
        The initial scaling.
    alpha : float
        A regularisation parameter.

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    BFGS1_product, invBFGS1_product, Broyden1_product, invBroyden1_product,
    Broyden2_product, invBroyden2_product
    """
    if Hessian_product is BFGS1_product:
        return BFGS1_product(u_vec, S_mat, Y_mat, scale)

    if Hessian_product is invBFGS1_product:
        return invBFGS1_product(u_vec, S_mat, Y_mat, scale)

    if Hessian_product is Broyden1_product:
        return Broyden1_product(mem, u_vec, S_mat, Y_mat, scale)

    if Hessian_product is invBroyden1_product:
        return invBroyden1_product(u_vec, S_mat, Y_mat, scale, alpha)

    if Hessian_product is Broyden2_product:
        return Broyden2_product(mem, u_vec, S_mat, Y_mat, scale)

    if Hessian_product is invBroyden2_product:
        return invBroyden2_product(u_vec, S_mat, Y_mat, scale, alpha)



"""Contains all functions that are used to calculate the next step.
"""
from numpy import abs, amax, c_, diag, dot, r_, sum, sqrt, tril, triu, zeros, savez
from numpy.linalg import cholesky, LinAlgError, norm, solve

from .hessian import hessian_wrapper, scaler_wrapper
from .history import history_wrapper


# PYTHON PORT ANNOTATION
# Renamed delinv to scaleinv as del is a keyword in Python.
# Removed parameter dim, as it was not used.
def inv_tGammPhiTPhi_product2(mem, u_vec, S_mat, Y_mat, scaleinv, omega):
    """TODO

    Parameters
    ----------
    mem : int
        TODO
    u_vec : array_like
        TODO
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.
    scaleinv : float
        The initial scaling.
    omega : float
        TODO

    Returns
    -------
    array_like
        TODO

    See Also
    --------
    TRFBGS1
    """
    # Build the pieces:
    tau = scaleinv + omega
    # PYTHON PORT ANNOTATION
    # Removed all variables for transposed matrices, that were only used once.
    YtY_mat = dot(Y_mat.T, Y_mat)
    St_mat = S_mat.T
    StS_mat = dot(St_mat, S_mat)
    StY_mat = dot(St_mat, Y_mat)
    D_mat = diag(diag(StY_mat))
    Rbar_mat = triu(StY_mat, 0)
    L_mat = tril(StY_mat, -1)
    K_mat = YtY_mat + tau*D_mat
    try:
        K_mat = cholesky(K_mat).T
    except LinAlgError:
        print('Not positive definite')
    Kt_mat = K_mat.T
    E_mat = scaleinv*Rbar_mat - omega*L_mat
    Et_mat = E_mat.T
    B_mat = omega*scaleinv*StS_mat\
            + dot(E_mat, solve(K_mat, solve(Kt_mat, Et_mat)))

    # print('--- in inv_tGammPhiTPhi_product2:29 ---')
    # print('S_mat =\n', S_mat)
    # print('StS_mat =\n', StS_mat)
    # print('B_mat =\n', B_mat)
    # print('--- out inv_tGammPhiTPhi_product2:33 ---')

    try:
        B_mat = cholesky(B_mat).T
    except LinAlgError:
        print('Not positive definite')
    Bt_mat = B_mat.T

    # print('--- in inv_tGammPhiTPhi_product2:34 ---')
    # print('K_mat =\n', K_mat)
    # print('E_Mat =\n', E_mat)
    # print('B_mat =\n', B_mat)
    # print('--- out inv_tGammPhiTPhi_product2:38 ---')

    # Execute inverse multiplication:
    u01_vec = u_vec[0:mem]
    u02_vec = u_vec[mem:2*mem]

    u11_vec = solve(Bt_mat, u01_vec)
    u11_vec = solve(B_mat, u11_vec)

    u21_vec = dot(Et_mat, u11_vec)
    u21_vec = solve(Kt_mat, u21_vec)
    u21_vec = solve(K_mat, u21_vec)

    u12_vec = solve(Kt_mat, u02_vec)
    u12_vec = solve(K_mat, u12_vec)

    u22_vec = dot(E_mat, u12_vec)
    u22_vec = solve(Bt_mat, u22_vec)
    u22_vec = solve(B_mat, u22_vec)

    u32_vec = dot(Et_mat, u22_vec)
    u32_vec = solve(Kt_mat, u32_vec)
    u32_vec = solve(K_mat, u32_vec)

    v_vec = u11_vec - u22_vec
    v_vec = r_[v_vec, u32_vec - u21_vec - u12_vec]

    return v_vec


# PYTHON PORT ANNOTATION
# Renamed del to scale, as del is a keyword in Python.
# Removed parameter dim, as it was not needed.
def TRBFGS1(mem, u_vec, ngrad, S_mat, Y_mat, tr, scale):
    """TODO

    Parameters
    ----------
    mem : int
        How many steps to look back.
    u_vec : array_like
        The negative gradient vector.
    ngrad : float
        The norm of the current gradient.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.
    tr : float
        The radius of the trust region.
    scale : float
        The initial scaling.

    Returns
    -------
    d_vec : array_like
        The next step.

    See Also
    --------
    Explicit_TR : Calling function.
    inv_tGammPhiTPhi_product2 : Used in this function.
    """
    # PYTHON PORT ANNOTATION
    # Removed temp_mat, if it was only used once before reassignment.
    scaleinv = 1./scale
    Phi_mat = c_[scaleinv*S_mat, Y_mat]
    v0_vec = dot(Phi_mat.T, u_vec)
    # OPT Remove temp_mat
    temp_mat = S_mat.T
    StS_mat = dot(temp_mat, S_mat)
    StY_mat = dot(temp_mat, Y_mat)
    # PYTHON PORT ANNOTATION
    # Replaced spdiags with diag, as the sparsity of the matrix was
    # never used.
    # TODO(benedikt) Check that sparsity is never used.
    D_mat = diag(diag(StY_mat))
    YtY_mat = dot(Y_mat.T, Y_mat)
    temp_mat = scaleinv*StY_mat
    Psi_mat = r_[c_[scaleinv**2*StS_mat, temp_mat], c_[temp_mat.T, YtY_mat]]
    temp_mat = tril(StY_mat, -1)
    Gamma_mat = r_[c_[scaleinv*StS_mat, temp_mat], c_[temp_mat.T, -D_mat]]

    # savez('py_vars.npz', Phi_mat=Phi_mat, v0_vec=v0_vec,
    #       StS_mat=StS_mat, StY_mat=StY_mat, D_mat=D_mat,
    #       YtY_mat=YtY_mat, Psi_mat=Psi_mat, Gamma_mat=Gamma_mat,
    #       u_vec=u_vec)
    # print('--- in TRBFGS1:86 ---')
    # print('Psi_mat =\n', Psi_mat)
    # print('Gamma_mat =\n', Gamma_mat)
    # print('--- out TRBFGS1:88 ---')

    Maxit_TRsub = 12
    # Maxit_TRsub = 0
    Tol_TRsub = 1e-10
    Iter_TRsub = 0
    dmu = 1.
    mu_max = ngrad/tr
    mu0 = mu_max/2.
    while Iter_TRsub <= Maxit_TRsub and abs(dmu) > Tol_TRsub:
        Iter_TRsub += 1
        tau = scaleinv + mu0
        v1_vec = inv_tGammPhiTPhi_product2(mem, v0_vec, S_mat, Y_mat,
                                           scaleinv, mu0)
        v2_vec = dot(Gamma_mat, v1_vec)
        v2_vec = inv_tGammPhiTPhi_product2(mem, v2_vec, S_mat, Y_mat,
                                           scaleinv, mu0)

        # print('--- in TRBFGS1:106 ---')
        # print('Iter_TRsub =', Iter_TRsub)
        # print('v1_vec =\n', v1_vec)
        # print('v2_vec =\nf', v2_vec)
        # print('--- out TRBFGS1:109 ---')

        temp = dot(v0_vec, v1_vec)
        sigma = dot(v1_vec, dot(Psi_mat, v1_vec)) + 2*temp + ngrad**2
        # savez('py_vars.npz', v1_vec=v1_vec, v2_vec=v2_vec,
        #       v0_vec=v0_vec, S_mat=S_mat, Y_mat=Y_mat, scaleinv=scaleinv,
        #       mu0=mu0, sigma=sigma)
        if sigma < 0 or Iter_TRsub == Maxit_TRsub:
            sigma = 0.

        # PYTHON PORT ANNOTATION
        # Removed sqrtsigma, as it was only used once.
        temp = sqrt(sigma) - tau*tr
        Delta = sigma + tau*(dot(v1_vec, dot(Psi_mat, v2_vec))
                             + dot(v0_vec, v2_vec))
        dmu = (sigma/Delta)*(temp/tr)
        mu1 = mu0 + dmu
        if mu1 >= 0:
            if mu1 >= mu_max:
                if mu1 == mu_max:
                    print('Upper bound on mu is wrong.')
                else:
                    mu0 = mu_max
            else:
                mu0 = mu1
        else:
            mu0 = .2*mu0
        # savez('py_vars.npz', temp=temp, Delta=Delta, dmu=dmu, mu1=mu1, mu0=mu0)

    # savez('py_vars.npz', dmu=dmu, mu0=mu0, mu_max=mu_max,
    #       v1_vec=v1_vec, tau=tau)
    temp = dot(Phi_mat, v1_vec)
    d_vec = (u_vec + temp)/tau

    return d_vec


# PYTHON PORT ANNOTATION
# Renamed iter to it, as iter is a keyword in Python.
def Step_generator(Step_finder, Scaler, History, Hessian_product,
                   invHessian_product, it, mem, c, beta, gamma, x_mat,
                   gradf_mat, ngradf_vec, stepsize_vec, tr, trstep,
                   alpha_0, verbose):
    """Calculates the direction of the next step and it's inital length.

    Parameters
    ----------
    Step_finder : function
        TODO
    Scaler : function
        TODO
    History : function
        TODO
    Hessian_product : function
        TODO
    invHessian_product : function
        TODO
    it : int
        The current iteration.
    mem : int
        TODO
    c : float
        TODO
    beta : float
        TODO
    gamma : float
        TODO
    x_mat : array_like
        TODO
    gradf_mat : array_like
        TODO
    ngradf_vec : array_like
        TODO
    stepsize_vec : array_like
        TODO
    tr : float
        TODO
    trstep : int
        TODO
    alpha_0 : float
        TODO
    verbose : bool
        TODO

    Returns
    -------
    d_vec : array_like
        The next step.
    """
    dim = x_mat.shape[0]
    ######################################################################
    # choose search direction and length:
    ######################################################################
    if mem > 0:
        S_mat = zeros((dim, mem))
        Y_mat = zeros((dim, mem))
        # print('--- in Step_generator:23 ---')
        # print('x_mat[:, -mem-2:-1] =', x_mat[:, -mem-2:-1])
        # print('gradf_mat[:, -mem-2:-1] =',
        #       gradf_mat[:, -mem-2:-1])
        # print('--- out Step_generator:27 ---')
        S_mat, Y_mat = history_wrapper(History, mem+1,
                                       x_mat[:, -mem-2:-1],
                                       gradf_mat[:, -mem-2:-1])
        delta = scaler_wrapper(Scaler, mem, S_mat, Y_mat, alpha_0)
        alpha = alpha_0
        # print('--- in Step_generator:33 ---')
        # print('S_mat =', S_mat)
        # print('Y_mat =', Y_mat)
        # print('delta =', delta)
        # print('--- out Step_generator:37 ---')
        d_vec = step_wrapper(Step_finder, Hessian_product,
                             invHessian_product, mem,
                             gradf_mat[:, -2],
                             ngradf_vec[it], S_mat, Y_mat, tr,
                             delta, alpha, verbose)
        stepsize_vec[it] = norm(d_vec)
        # savez('py_vars.npz', gradf_mat=gradf_mat, ngradf_vec=ngradf_vec,
        #       tr=tr, alpha=alpha)
    else:
        # steepest descent direction
        delta = 0
        alpha = alpha_0
        # if we get to this option, the step MUST have been accepted
        # and iter updated
        d_vec = -stepsize_vec[it]*gradf_mat[:, -2]/ngradf_vec[it]

    if abs(stepsize_vec[it] - tr) < 1e-11:
        trstep += 1

    DDf = dot(gradf_mat[:, -2], d_vec)
    while DDf > 0:
        if verbose:
            print('The Step generator yields a direction of nondescent. ')

        # shrink memory and try again
        if mem > 2:
            S_mat = S_mat[:, 1:mem]
            Y_mat = Y_mat[:, 1:mem]
            delta = scaler_wrapper(Scaler, mem-1, S_mat, Y_mat, alpha_0)
            alpha = alpha_0
            d_vec = step_wrapper(Step_finder, Hessian_product,
                                 invHessian_product, mem-1,
                                 gradf_mat[:, -2],
                                 ngradf_vec[it], S_mat, Y_mat,
                                 tr, delta, alpha)
        else:
            # steepest descent
            d_vec = -gamma*stepsize_vec[it]*gradf_mat[:, -2]\
                    / ngradf_vec[it]

        mem -= 1
        DDf = dot(gradf_mat[:, -2], d_vec)

    ##################################################
    # end choose search direction and length:
    ##################################################
    cDDf = c*DDf
    bDDf = beta*DDf

    return delta, alpha, d_vec, stepsize_vec, trstep, DDf, cDDf, bDDf, mem


# PYTHON PORT ANNOTATION
# Renamed iter to it as iter is a keyword in Python.
def Step_length_update(History, Hessian_product, update_conditions,
                       mem, it, x_mat, f_vec,
                       gradf_mat, d_vec, mu_vec, stepsize_vec, tr,
                       delta, alpha, eta1, eta2, eta3, cDDf, DDfnew,
                       bDDf, orthog_TOL, lower_step_bound, gamma):
    """Trust region adjustment routine.

    Parameters
    ----------
    History : function
        TODO
    Hessian_product : function
        TODO
    update_conditions : str
        TODO
    mem : int
        TODO
    it : int
        The current iteration.
    x_mat : array_like
        TODO
    f_vec : array_like
        TODO
    gradf_mat : array_like
        TODO
    d_vec : array_like
        TODO
    mu_vec : array_like
        TODO
    stepsize_vec : array_like
        TODO
    tr : float
        TODO
    delta : float
        TODO
    alpha : float
        TODO
    eta1 : float
        TODO
    eta2 : float
        TODO
    eta3 : float
        TODO
    cDDf : float
        TODO
    DDfnew : float
        TODO
    bDDf : float
        TODO
    orthog_TOL : float
        TODO
    lower_step_bound : float
        TODO
    gamma : float
        TODO

    Returns
    -------
    tr : float
        TODO
    mu_vec : array_like
        TODO
    cDDf : float
        TODO
    lower_step_bound : float
        TODO
    d_vec : array_like
        TODO
    stepsize_vec : array_like
        TODO
    accept : int
        TODO
    """
    # print('--- in Step_length_update:86 ---')
    # print('update_conditions =', update_conditions)
    # print('mem =', mem)
    # print('--- out Step_length_update:89 ---')
    if update_conditions == 'Exact':
        ##################################################
        # exact line search via bisection
        ##################################################
        if mu_vec[it] == 0:
            tr = 1e+15
            lower_step_bound = 0
            mu_vec[it] = 1

        min_d = lower_step_bound
        max_d = tr
        nd = stepsize_vec[it]
        if abs(DDfnew)/stepsize_vec[it] <= orthog_TOL:
            accept = 1
        else:
            accept = -2
            if f_vec[it+1] <= f_vec[it] and DDfnew/nd < -orthog_TOL:
                # step too short:
                if abs(nd-max_d)/(max_d-min_d) < 1e-2:
                    max_d = 1e+7*max_d
                    tr = max_d
                    d_vec = 2*d_vec
                else:
                    # increase step length by one half the interval to
                    # the outer bound
                    d_vec = min(nd+max_d/2., 10.*nd)*d_vec/nd
                min_d = nd
            elif f_vec[it+1] <= f_vec[it] and DDfnew/nd > orthog_TOL:
                # step too long, but not by too much:
                # decrease step length by one half the interval to the
                # inner bound
                d_vec = (nd+min_d)/2.*(d_vec/nd)
                max_d = nd
                tr = max_d
            elif f_vec[it+1] > f_vec[it]:
                # step too long, by a lot:
                # decrease step length by one half the current steplength
                max_d = nd
                tr = max_d
                d_vec = .5*(min_d+nd)*(d_vec/nd)

        lower_step_bound = min_d
        stepsize_vec[it] = norm(d_vec)

    # do backtracking linesearch/ adjust the trust region
    else:
        if mem > 0:
            S_mat, Y_mat = history_wrapper(History, mem+1,
                                           x_mat[:, -mem-2:-1],
                                           gradf_mat[:, -mem-2:-1])

        ##################################################
        # Adjust trust region:
        ##################################################
        # actual change/predicted change
        # look-back - this is a type of nonmonotone trust region adjustment
        p = max(mem-1, 1)
        pind = max(it-p, 0)
        # PYTHON PORT ANNOTATION
        # Added initialisation of pred here, so that it always exists.
        pred = 0.
        if mem > 0:
            pred = dot(gradf_mat[:, -2], d_vec) \
                   + .5*dot(d_vec, hessian_wrapper(Hessian_product, mem,
                                                   d_vec, S_mat, Y_mat, delta,
                                                   alpha))
            actual = f_vec[it+1] - f_vec[it]
            mu_vec[it] = actual/pred
            if abs(pred-actual)/abs(pred) > 1e+5:
                print('Warning: step length may be too small -- samsara results might be at numerical accuracy.')
                print('Try increasing your step length tolerance.')
        else:
            mu_vec[it] = (f_vec[it+1]-f_vec[it])\
                               / dot(gradf_mat[:, -2], d_vec)

        # savez('py_vars.npz', mem=mem, mu_vec=mu_vec, pred=pred,
        #       stepsize_vec=stepsize_vec, tr=tr, pind=pind)
        if mem > 0:
            if mu_vec[it] <= 0 or pred > 0:
                tr = min(.1*stepsize_vec[it-1], .1*tr)
                accept = -1
            elif sum(mu_vec[pind+1:it+1])/(it-pind) >= eta1:
                tr = min(5.*tr, 1e+15)
                accept = 1
            elif sum(mu_vec[pind+1:it+1])/(it-pind) >= eta2:
                tr = min(2.*tr, 1e+15)
                accept = 1
            elif sum(mu_vec[pind+1:it+1])/(it-pind) > eta3:
                # Nocedal & Wright would suggest leaving tr the same
                tr = 1.2*tr
                accept = 1
            else:
                tr = min(stepsize_vec[it], .25*tr)
                if f_vec[it+1] <= amax(f_vec[max(0, it-p):it+1])\
                   + cDDf and DDfnew >= bDDf:
                    accept = 1
                elif f_vec[it+1] <= amax(f_vec[max(0, it-p):it+1])\
                     and stepsize_vec[it] < 2.*stepsize_vec[it-1]:
                    # still got decrease, so will keep the
                    # information, but not advance the point
                    accept = 0
                else:
                    # no decrease and stepsize might be too small,
                    # shrink memory and don't keep info
                    accept = -1
        else:
            max_d = tr
            nd = stepsize_vec[it]
            if lower_step_bound >= max_d:
                min_d = 0
            else:
                min_d = lower_step_bound
            # weakness of this method.  In general don't know what max_d
            # should be, and if you underestimate the routine will stagnate.
            if abs(nd-max_d)/(max_d-min_d) < 1e-2:
                max_d *= 1e+7
                tr = max_d

            ##################################################
            # nonmonotone bisection line search
            ##################################################
            p = max(mem, 1)
            # TODO(benedikt) DELETE ME
            # print('--- start Step_length_update:203 ---')
            # print('f_vec =', f_vec)
            # print('p =', p)
            # print('it =', it)
            # print('f_vec[max(0, it-p):it] =', f_vec[max(0, it-p):it])
            # print('--- end Step_length_update:208 ---')
            if f_vec[it+1] <= amax(f_vec[max(0, it-p):it+1])+cDDf\
               and DDfnew >= bDDf:
                # Wolfe conditions satisfied, do not change step.
                accept = 1
            else:
                accept = -2
                if f_vec[it+1] <= amax(f_vec[max(0, it-p):it+1])+cDDf\
                   and DDfnew < bDDf:
                    # probably underestimating the upper bound
                    if (max_d-min_d)/max_d <= 1e-2:
                        max_d *= 10.
                        tr = max_d
                    # step too short:
                    # increase step length to the outer bound
                    d_vec = min(.1*nd + .9*max_d, 100.*nd)*d_vec/nd
                    min_d = nd
                    accept = 1
                elif f_vec[it+1] <= amax(f_vec[max(0, it-p):it+1])\
                     and DDfnew >= bDDf:
                    # step too long, but not by too much:
                    # decrease step length by one half the
                    # interval to the inner bound
                    d_vec = (nd+min_d)/2.*(d_vec/nd)
                    max_d = nd
                    tr = max_d
                elif f_vec[it+1] >= f_vec[it]:
                    # step too long, by a lot:
                    # decrease step length by one half the current steplength
                    max_d = nd
                    d_vec = .5*(min_d+nd)*(d_vec/nd)
                    tr = max_d
                    cDDf *= gamma
            lower_step_bound = min_d
            stepsize_vec[it] = norm(d_vec)
        ##################################################
        # end trust region adjust
        ##################################################

    # savez('py_vars.npz', tr=tr, mu_vec=mu_vec, cDDf=cDDf,
    #       lower_step_bound=lower_step_bound, d_vec=d_vec,
    #       stepsize_vec=stepsize_vec, accept=accept)

    return tr, mu_vec, cDDf, lower_step_bound, d_vec, stepsize_vec, accept


# PYTHON PORT ANNOTATION
# Renamed del to scale, as del is a keyword in Python.
def Dogleg_QN(Hessian_product, invHessian_product, mem, grad_vec,
              ngrad, S_mat, Y_mat, tr, scale, alpha):
    """Stable Quasi-Newton Dogleg.

    Dogleg subroutine for line search methods.

    Parameters
    ----------
    Hessian_product : function
        The dogleg method.
    invHessian_product : function
        The quasi newton method.
    mem : int
        How far to look back.
    grad_vec : array_like
        The current gradient.
    ngrad : float
        The norm of the current gradient.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matix of gradients.
    tr : float
        The radius of the trust region.
    scale : float
        The initial scaling.
    alpha : float
        A regularisation parameter.

    Returns
    -------
    d_vec : array_like
        The next step. Equals the step calculated with the quasi
    newton method, if that step already is within the trust region.

    Notes
    -----
    Solves the following problem:
    """
    dn = hessian_wrapper(invHessian_product, mem, -grad_vec,
                         S_mat, Y_mat, scale, alpha)
    # print('--- in Dogleg_QN:22 ---')
    # print('dn =', dn)
    # print('--- out Dogleg_QN:24 ---')
    if norm(dn) <= tr:   # to turn on  set <= 1e+15
        ddl = dn
    else:
        # initialization
        u = -grad_vec/ngrad
        v = hessian_wrapper(Hessian_product, mem, u, S_mat,
                            Y_mat, scale, alpha)
        w = dot(u, v)
        # predicted change at constrained solution
        b = -tr*ngrad + .5*(tr**2)*w
        # t_star solves min{-t ||grad_f||
        # + 1/2*t^2* grad_f'*H*grad_f/||grad_f||^2}
        # => t_star = ||grad_f||^3/grad_f'*H*grad_f
        # from first order necessary conditions for optimality
        t_star = ngrad/w
        if t_star < 0 or t_star > tr:
            if b > 0:
                t = 0
            else:
                t = tr
        else:
            # predicted change at unconstrained solution
            c = -.5*(ngrad**2)/w
            # only want the t that solves
            # min{-t ||grad_f|| + 1/2*t^2* grad_f'*H*grad_f/||grad_f||^2}
            # on [0, tr]
            if c < 0 or b < 0:
                if c < b:
                    t = t_star
                else:
                    t = tr
            else:
                t = 0

        dsd = t*u

        # PYTHON PORT ANNOTATION
        # Replaced norm(dsd-dn)**2 with dot(dsd-dn, dsd-dn).
        # TODO(benedikt) add in extra variables for dsd-dn and dot(dsd-dn, dsd)
        tmp = dot(dsd-dn, dsd-dn)
        disc = dot(dsd-dn, dsd)**2 + tmp*(tr**2 - dot(dsd, dsd))
        lamb = max((sqrt(disc) - dot(dn-dsd, dsd))/tmp, 0)
        ddl = lamb*dn + (1-lamb)*dsd

    return ddl


# PYTHON PORT ANNOTATION
# Renamed del to scale, as del is a keyword in Python.
def Explicit_TR(invHessian_product, mem, grad_vec, ngrad,
                S_mat, Y_mat, tr, scale, alpha, verbose):
    """Stable Explicit Trust Region.

    Stable Explicit Trust Region for Symmetric PSD matrix secant methods.

    Parameters
    ----------
    invHessian_product : function
        The quasi newton method.
    mem : int
        How many steps to look back.
    grad_vec : array_like
        The current gradient.
    ngrad : float
        The norm of the current gradient.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.
    tr : float
        The radius of the trust region.
    scale : float
        The initial scaling.
    alpha : float
        A regularisation parameter.
    verbose : bool
        If additional output should be printed.

    Returns
    -------
    d_vec : array_like
        The next step.

    See Also
    --------
    TRBFGS1, Dogleg_QN
    """
    TRsub = TRBFGS1
    dn_vec = hessian_wrapper(invHessian_product, mem, -grad_vec,
                             S_mat, Y_mat, scale, alpha)
    # print('--- in Explicit_TR:84 ---')
    # print('dn_vec =\n', dn_vec)
    # print('--- out Explicit_TR:86 ---')
    if norm(dn_vec) <= tr:   # to turn on  set <= 1e+15
        d_vec = dn_vec
    else:
        if verbose:
            print('in trust region subproblem')

        d_vec = TRsub(mem, -grad_vec, ngrad, S_mat, Y_mat, tr, scale)

    # savez('py_vars.npz', dn_vec=dn_vec, d_vec=d_vec)
    return d_vec


# PYTHON PORT ANNOTATION
# Renamed del to scale, as del is a keyword in Python.
def step_wrapper(step_finder, Hessian_product, invHessian_product, mem,
                 grad_vec, ngrad, S_mat, Y_mat, tr, scale, alpha,
                 verbose):
    """A wrapper for all step finding functions.

    The wrapper provides a unified function call for all step finders.

    Parameters
    ----------
    step_finder : function
        The step finder to use (Dogleg_QN, Explicit_TR).
    Hessian_product : function
        The dogleg method for Dogleg_QN.
    invHessian_product : function
        The quasi newton method.
    mem : int
        How many steps to look back.
    grad_vec : array_like
        The current gradient.
    ngrad : float
        The norm of the current gradient.
    S_mat : array_like
        The matrix of steps.
    Y_mat : array_like
        The matrix of gradients.
    tr : float
        The radius of the trust region.
    scale : float
        The initial scaling.
    alpha : float
        A regularisation parameter.
    verbose : bool
        If additional output should be printed.

    Returns
    -------
    d_vec - array_like
        The next step as calculated by the `step_finder`.

    See Also
    --------
    Dogleg_QN, Explicit_TR
    """
    if step_finder is Explicit_TR:
        return Explicit_TR(invHessian_product, mem, grad_vec, ngrad,
                           S_mat, Y_mat, tr, scale, alpha, verbose)

    if step_finder is Dogleg_QN:
        return Dogleg_QN(Hessian_product, invHessian_product, mem,
                         grad_vec, ngrad, S_mat, Y_mat, tr, scale, alpha)



from numpy import dot
from numpy.linalg import norm


# PYTHON PORT ANNOTATION
# Renamed iter to it, as iter is a keyword in Python.
# Removed parameter maxmem, as it is not needed in Python.
def Memory_update(x_mat, f_vec, gradf_mat, ngradf_vec, stepsize_vec,
                  mem, psd, it, accept_step):
    """Memory update routine.

    Parameters
    ----------
    x_mat : array_like
        TODO
    f_vec : array_like
        TODO
    gradf_mat : array_like
        TODO
    ngradf_vec : array_like
        TODO
    stepsize_vec : array_like
        TODO
    mem : int
        TODO
    psd : int
        TODO
    it : int
        The current iteration.
    accept_step : int
        TODO

    Returns
    -------
    x_mat : array_like
        TODO
    f_vec : array_like
        TODO
    gradf_mat : array_like
        TODO
    ngradf_vec : array_like
        TODO
    stepsize_vec : array_like
        TODO
    mem : int
        TODO
    """
    # print('--- in Memory_update:12 ---')
    # print('mem =', mem)
    # print('psd =', psd)
    # print('accept_step =', accept_step)
    # print('--- out Memory_update:17---')
    ######################################################################
    # Update the memory
    ######################################################################
    if psd == 1:
        # extra check that sty>0 for BFGS-based methods
        s_vec = x_mat[:, -1] - x_mat[:, -2]
        y_vec = gradf_mat[:, -1] - gradf_mat[:, -2]
        sty = dot(s_vec, y_vec)
        sty_tol = norm(y_vec)*norm(s_vec)*1e-12

        # print('--- in Memory_update:28 ---')
        # print('sty =', sty)
        # print('sty_tol =', sty_tol)
        # print('--- out Memory_update:31 ---')
        if sty <= sty_tol:
            # lost positive defniniteness:  delete old memory elements
            if accept_step == 1:
                # still made progress, so keep new point and reset history
                mem = 0
                x_mat[:, -2] = x_mat[:, -1]
                gradf_mat[:, -2] = gradf_mat[:, -1]
                # PYTHON PORT ANNOTATION
                # Replaced sqrt(dot(...)) with norm(...).
                ngradf_vec[it+1] = norm(gradf_mat[:, -1])
            elif accept_step == 0:
                # did not make progress, reset history starting from
                # current point
                mem = 0
                ngradf_vec[it+1] = ngradf_vec[it]

        else:
            if accept_step == 0:
                # No progress, but keep new information.
                # Recenter the steps: keep the current step the center
                #    and move the new information back in the memory
                f_vec[it], f_vec[it+1] = \
                    f_vec[it+1], f_vec[it]
                ngradf_vec[it+1] = ngradf_vec[it]
                # PYTHON PORT ANNOTATION
                # Replaced sqrt(dot(...)) with norm(...).
                ngradf_vec[it] = norm(gradf_mat[:, -1])

                # update memory
                mem = min(mem+1, x_mat.shape[1]-2)
                # TODO(benedikt) Check if values are correctly adjusted to 0 based indexing.
                x_mat[:, 0:-2] = x_mat[:, 1:-1]
                x_mat[:, -2] = x_mat[:, -1]
                # TODO(benedikt) Check if values are correctly adjusted to 0 based indexing.
                gradf_mat[:, 0:-2] = gradf_mat[:, 1:]
                gradf_mat[:, -2] = gradf_mat[:, -1]
                # PYTHON PORT ANNOTATION
                # Replaced sqrt(dot(...)) with norm(...).
                ngradf_vec[it+1] = norm(gradf_mat[:, -1])
            elif accept_step == -1:
                # No progress, don't take the step and shrink memory.
                # update memory
                mem = max(mem-1, 0)
                ngradf_vec[it+1] = ngradf_vec[it]
            elif accept_step == -2:
                # did not make progress, reset history starting from
                # current point
                mem = 0
                ngradf_vec[it+1] = ngradf_vec[it]
            elif accept_step == 1:
                # step accepted --> model working
                # take the proposed step and update memory
                mem = min(mem+1, x_mat.shape[1]-2)
                x_mat[:, 0:-2] = x_mat[:, 1:-1]
                x_mat[:, -2] = x_mat[:, -1]
                gradf_mat[:, 0:-2] = gradf_mat[:, 1:-1]
                gradf_mat[:, -2] = gradf_mat[:, -1]
                # PYTHON PORT ANNOTATION
                # Replaced sqrt(dot(...)) with norm(...).
                ngradf_vec[it+1] = norm(gradf_mat[:, -1])

    elif accept_step == 0:
        # step not acctepted --> past information
        # is no good, but don't throw away new information.
        # Recenter the steps: keep the current step the center
        #    and move the new information back in the memory
        f_vec[it], f_vec[it+1] = f_vec[it+1], f_vec[it]
        ngradf_vec[it+1] = ngradf_vec[it]
        # PYTHON PORT ANNOTATION
        # Replaced sqrt(dot(...)) with norm(...).
        ngradf_vec[it] = norm(gradf_mat[:, -1])

        # update memory
        mem = min(mem+1, x_mat.shape[1]-2)
        x_mat[:, 0:-2] = x_mat[:, 1:-1]
        x_mat[:, -2] = x_mat[:, -1]
        gradf_mat[:, 0:-2] = gradf_mat[:, 1:-1]
        gradf_mat[:, -2] = gradf_mat[:, -1]
        # PYTHON PORT ANNOTATION
        # Replaced sqrt(dot(...)) with norm(...).
        ngradf_vec[it+1] = norm(gradf_mat[:, -1])
    elif accept_step == -1:
        # No progress, don't take the step and shrink memory.
        # update memory
        mem = max(mem-1, 0)
        ngradf_vec[it+1] = ngradf_vec[it]
    elif accept_step == -2:
        # did not make progress, reset history starting from
        # current point
        mem = 0
        ngradf_vec[it+1] = ngradf_vec[it]
    elif accept_step == 1:
        # model working
        # take the proposed step and update memory
        mem = min(mem+1, x_mat.shape[1]-2)

    stepsize_vec[it+1] = stepsize_vec[it]

    # print('--- in Memory_update:128 ---')
    # print('mem =', mem)
    # print('maxmem =', maxmem)
    # print('--- out Memory_update:131---')

    return x_mat, f_vec, gradf_mat, ngradf_vec, stepsize_vec, mem
